import mariadb
import os
import time

from Mine.Block import Block
from Mine.Transaction import Transaction

class Database():
    SQL_CREATE_DATABASE = [
        """
        CREATE TABLE IF NOT EXISTS mine.blocks(
            sqlId INT PRIMARY KEY AUTO_INCREMENT,
            blockId VARCHAR(37) UNIQUE NOT NULL,
            blockHash VARCHAR(64) UNIQUE NOT NULL,
            previousHash VARCHAR(64) UNIQUE NOT NULL,
            timestamp DATETIME NOT NULL,
            height INT UNIQUE NOT NULL,
            nonce INT NOT NULL
        );
        """,
        """
        CREATE TABLE IF NOT EXISTS mine.cards(
            sqlId INT PRIMARY KEY AUTO_INCREMENT,
            blockId VARCHAR(37) UNIQUE NOT NULL,
            cardId VARCHAR(37) UNIQUE NOT NULL,
            pageId INT NOT NULL,
            FOREIGN KEY (blockId) REFERENCES blocks(blockId)
        );
        """,
        """
        CREATE TABLE IF NOT EXISTS mine.transactions(
            sqlId INT PRIMARY KEY AUTO_INCREMENT,
            blockId VARCHAR(37) NOT NULL,
            cardId VARCHAR(37) UNIQUE NOT NULL,
            transactionId VARCHAR(37) UNIQUE NOT NULL,
            timestamp DATETIME NOT NULL,
            sender VARCHAR(128) NOT NULL,
            receiver VARCHAR(128) NOT NULL,
            signature VARCHAR(128) NOT NULL,
            isPending BOOLEAN NOT NULL DEFAULT true,
            isAbandoned BOOLEAN NOT NULL DEFAULT false,
            FOREIGN KEY (blockId) REFERENCES blocks(blockId),
            FOREIGN KEY (cardId) REFERENCES cards(cardId)
        );
        """,
        """
        CREATE TABLE IF NOT EXISTS mine.peers(
            sqlId INT PRIMARY KEY AUTO_INCREMENT,
            peerId VARCHAR(37) UNIQUE NOT NULL,
            baseUrl VARCHAR(128) UNIQUE NOT NULL,
            isUp BOOLEAN NOT NULL,
            downCount INT DEFAULT 0,
            lastTry TIMESTAMP DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP
        );
        """
        ]

    SQL_GET_BLOCK_BY_ID = """
        SELECT *
        FROM blocks
        WHERE blockId = '{}';
        """

    SQL_GET_LAST_BLOCK = """
        SELECT *
        FROM blocks
        ORDER BY timestamp DESC
        LIMIT 1;
        """

    SQL_GET_CARD = """
        SELECT *
        FROM cards
        WHERE cardId = '{}';
        """

    SQL_GET_CARD_OWNER = """
        SELECT receiver
        FROM transactions
        WHERE cardId = '{}'
        AND isPending = False
        ORDER BY timestamp DESC
        LIMIT 1;
        """

    SQL_GET_PENDING_TRANSACTIONS = """
        SELECT *
        FROM transactions
        WHERE isPending = True
        ORDER BY timestamp ASC
        LIMIT {};
        """

    SQL_GET_TRANSACTION_BY_ID = """
        SELECT *
        FROM transactions
        WHERE transactionId == '{}'
        """

    SQL_INSERT_BLOCK = """
        INSERT INTO blocks (blockId, blockHash, previousHash, timestamp, height, nonce)
        VALUES ('{}', '{}', '{}', '{}', {}, {});
        """

    SQL_INSERT_TRANSACTION = """
        INSERT INTO transactions (
            transactionId,
            timestamp,
            sender,
            receiver,
            signature,
            blockId,
            cardId
            )
        VALUES ({}, {}, {}, {}, {});
        """

    def __init__(self):
        delay = 2
        while True:
            try:
                self.conn = mariadb.connect(
                    host = os.getenv('MINE_DB_HOST', "mariadb"),
                    port = os.getenv('MINE_DB_PORT', 3306),
                    user = os.getenv('MINE_DB_USER', None),
                    password = os.getenv('MINE_DB_PASSWORD', None),
                    database = 'mine'
                    )
                for each in self.SQL_CREATE_DATABASE:
                    cursor = self.conn.cursor().execute(each)
            except mariadb.Error as e:
                time.sleep(delay := delay**2)
                continue
            break

    def get_block_by_id(self, blockId):
        return Block(
            data=self.conn.cursor().execute(
                self.SQL_GET_BLOCK_BY_ID.format(blockId)
                )
            )

    def get_last_block(self):
        data=self.conn.cursor().execute(self.SQL_GET_LAST_BLOCK)
        return Block(data=data) if data else None

    def get_card(self, cardId):
        return Card(
            data=self.conn.cursor().execute(self.SQL_GET_CARD.format(cardId))
            )

    def get_card_owner(self, cardId):
        return self.conn.cursor().execute(
            self.SQL_GET_CARD_OWNER.format(cardId)
            )

    def get_pending_transactions(self, limit=32768):
        pendingTransactions = self.conn.cursor().execute(
            self.SQL_GET_PENDING_TRANSACTIONS.format(limit)
            )
        return [
            Transaction(
                data=each
                ) for each in pendingTransactions
            ] if pendingTransactions else []

    def get_transaction(self, transactionId):
        return Transaction(
            data=self.conn.cursor().execute(self.SQL_GET_TRANSACTION.format(transactionId))
            )

    def insert_block(self, block):
        self.conn.cursor().execute(self.SQL_INSERT_BLOCK.format(
            str(block.blockId),
            block.blockHash.hexdigest(),
            block.previousHash,
            str(block.timestamp),
            block.height,
            block.nonce
            ))

    def insert_transaction(self, transaction):
        return self.conn.cursor().execute(self.SQL_INSERT_TRANSACTION.format(
            str(transaction.transactionId),
            str(transaction.timestamp),
            transaction.sender,
            transaction.receiver,
            transaction.signature,
            str(transaction.cardId)
            ))
