import flask
import json
import os

from cryptography.hazmat.primitives.asymmetric import rsa

from Mine.Block import Block
from Mine.Card import Card
from Mine.Database import Database
from Mine.Transaction import Transaction

BLOCK_TRANSACTION_LIMIT = int(os.getenv('BLOCK_TRANSACTION_LIMIT', 32768))
DATA_PATH = os.getenv('DATA_PATH', '/var/lib/wikideck/blocks')
DIFFICULTY_REQUIREMENT = int(os.getenv('DIFFICULTY_REQUIREMENT', 0))

mine = flask.Blueprint("mine", __name__)
db = Database()
privateKey = rsa.generate_private_key(
    public_exponent=65537,
    key_size=4096
    )

def save_block(block):
    with open(f"{DATA_PATH}/{block.blockId}.json", 'w') as f:
        f.write(str(block)) # Blocks are json strings.  This is the true block.
    db.insert_block(block)

if not db.get_last_block():
    # TODO: load blocks from files
    # TODO: try to get blocks from peers
    originBlock = Block(
        difficulty = DIFFICULTY_REQUIREMENT,
        )
    originBlock.mine(privateKey)
    originBlock.validate()
    save_block(originBlock)

@mine.get('/')
def index_get():
    # TODO: return a page to mine through the browser
    return "Hello world!", 200

###
# Retrieve blocks and block data.
# Returns a skeleton block to be mined by default.
# Queries for a specific block when given parameters.
###
@mine.get('/blocks')
def blocks_get():
    # TODO: block queries
    blockId = flask.request.args.get('blockId', None)
    lastBlock = db.get_last_block()
    return flask.jsonify(
        Block(
            previousHash = lastBlock.blockHash.hexdigest(),
            height = lastBlock.height + 1,
            difficulty = DIFFICULTY_REQUIREMENT,
            transactions = [
                Transaction(data=each) for each in db.get_pending_transactions(
                    BLOCK_TRANSACTION_LIMIT
                    )
                ]
            ).as_dict()
        )

###
# Submit a block to the chain.
# This method calls the Block().validate() method to validate the block schema.
# This method performs additional validations against the rest of the chain.
###
@mine.post('/blocks')
def blocks_post():
    try:
        newBlock = Block(data=request.get_json())
        newBlock.validate()
        previousBlock = db.get_last_block()
        if newBlock.previousHash != previousBlock.blockHash:
            raise Block.Invalid(
                f"Incorrect previous hash - should be {previousBlock.blockHash}."
                )
        if newBlock.timestamp <= previousBlock.timestamp:
            raise Block.Invalid(
                "Timestamp is later than previous block."
                )
        if newBlock.height != previousBlock.height + 1:
            raise Block.Invalid(
                f"Incorrect block height - should be {previousBlock.height + 1}."
                )
        if newBlock.difficulty < DIFFICULTY_REQUIREMENT:
            raise Block.Invalid(
                f"Incorrect difficulty - should be {DIFFICULTY_REQUIREMENT}."
                )
        if len(newBlock.transactions) == 0:
            raise Block.Invalid(
                "Block contains no transactions."
                )
        for transaction in newBlock.transactions:
            pendingTransaction = db.get_transaction(transaction.transactionId)
            if not pendingTransaction:
                raise Transaction.Invalid(
                    f"No matching pending transaction for {transaction.transactionId}."
                    )
            if not pendingTransaction.pending:
                raise Transaction.AlreadyFulfilled(
                    f"Transaction {transaction.transactionId} has already been fulfilled."
                    )
            if transaction.timestamp != pendingTransaction.timestamp:
                raise Transaction.Invalid(
                    f"Incorrect timestamp on {transaction.transactionId}."
                    )
            if transaction.cardId != pendingTransaction.cardId:
                raise Transaction.Invalid(
                    f"Incorrect cardId on {transaction.transactionId}."
                    )
            if transaction.sender != pendingTransaction.sender:
                raise Transaction.Invalid(
                    f"Incorrect sender on {transaction.transactionId}."
                    )
            if transaction.recipient != pendingTransaction.recipient:
                raise Transaction.Invalid(
                    f"Incorrect recipient on {transaction.transactionId}."
                    )
            if transaction.signature != pendingTransaction.signature:
                raise Transaction.Invalid(
                    f"Incorrect signature on {transaction.transactionId}."
                    )
        save_block(block)
        for transaction in newBlock.transactions:
            db.update_transactions_is_pending_false(transaction.transactionId)
        # TODO: update peers
        return flask.jsonify(block.asDict())
    except Transaction.Invalid as e:
        return e, e.statusCode
    except Transaction.AlreadyFulfilled as e:
        return e, e.statusCode
    except Card.Invalid as e:
        return e, e.statusCode
    except Block.Invalid as e:
        return e, e.statusCode

###
# Retrieve card data
###
@mine.get('/cards')
def cards_get():
    # TODO: render cards in html
    # TODO: query cards
    # TODO: get decks
    return 200

###
# Submit a transaction to be mined in a block.
# This method performs a number of validations on the submitted transaction and returns
# a status code result.
###
@mine.put('/transactions')
def transactions_put():
    try:
        newTransaction = Transaction(data=request.get_json())
        newTransaction.validate()
        if not get_card(newTransaction.cardId):
            raise Transaction.Invalid(
                f"Card {newTransaction.cardId} does not exist.",
                404
                )
        if newTransaction.sender != get_card_owner(newTransaction.cardId):
            raise Transaction.Unauthorized(
                f"{newTransaction.sender} does not own {newTransaction.cardId}."
                )
        insert_transaction(newTransaction)
        # TODO: update peers?
        return 200
    except mariadb.Error as e:
        return e, 500
    except Transaction.Unauthorized as e:
        return e, e.statusCode
    except Transaction.Invalid as e:
        return e, e.statusCode

###
# Retrieve a transaction.
###
@mine.get('/transactions')
def transactions_get():
    transactionId = request.args.get('transactionId', None)
    return get_transaction_by_id(transactionId) if transactionId else json.dumps(
        db.get_pending_transactions()
        )

@mine.post('/peers')
def peers_post():
    try:
        peer = Peer(data=request.get_json())
        peer.validate()
        # TODO: add peers to database
        return 200
    # TODO: error handling
    except Exception as e:
        print('crap!')

@mine.get('/peers')
def peers_get():
    # TODO: query peers
    return 200
